# SPDX-License-Identifier: MIT
# Copyright © 2023-2024 Dylan Baker

"""An implementation of a Result type."""

from __future__ import annotations
from functools import wraps
from typing import TYPE_CHECKING, TypeVar, ParamSpec, Generic
from dataclasses import dataclass

if TYPE_CHECKING:
    from typing import Awaitable, Callable

    from .maybe import Maybe

P = ParamSpec('P')
R = TypeVar('R')
T = TypeVar('T')
U = TypeVar('U')
E = TypeVar('E')
F = TypeVar('F')

__all__ = [
    'Error',
    'ErrorWrapper',
    'Result',
    'Success',
    'UnwrapError',
    'stop',
    'stop_async',
    'unwrap_result',
    'unwrap_result_async',
    'wrap_result',
    'wrap_result_async',
]


class UnwrapError(Exception):
    """Error thrown when unwrapping is invalid."""


class ErrorWrapper(Exception):
    """wraps non Exception Errors"""


class Propagation(Generic[E], Exception):
    """Uses exception handling to propagate up."""

    def __init__(self, err: E) -> None:
        super().__init__('Uncaught Propogation, did you forget to decorate '
                         'function with @simple_monads.result.stop?')
        self.err = err


class Result(Generic[T, E]):

    """Base Class for Option, do not directly instantiate"""

    @staticmethod
    def is_ok() -> bool:
        """Returns True if this is a Success otherwise False."""
        raise NotImplementedError()

    @staticmethod
    def is_err() -> bool:
        """Returns True if this is an Error otherwise False."""
        raise NotImplementedError()

    def unwrap(self, msg: str | None = None) -> T:
        """Get the held value or throw an Exception.

        :param msg: The error message, otherwise a default is used
        :raises UnwrapError: If this is an Err
        :return: The held value
        """
        raise NotImplementedError()

    def unwrap_or(self, fallback: T) -> T:
        """Return the held value or fallback if this is an Error.

        :param fallback: A value to use incase of Error
        :return: The held value or the fallback
        """
        raise NotImplementedError()

    def unwrap_or_else(self, fallback: Callable[[], T]) -> T:
        """Return the held value or the result of the fallback.

        :param fallback: A callable to generate a result if this in Error
        :return: Either the held value or the fallback value
        """
        raise NotImplementedError()

    async def unwrap_or_else_async(self, fallback: Callable[[], Awaitable[T]]) -> T:
        """Return the held value or the result of the fallback.

        >>> import asyncio
        >>> async def fb():
        ...     return 'bar'
        >>> asyncio.run(Success('foo').unwrap_or_else_async(fb))
        'foo'
        >>> asyncio.run(Error(5).unwrap_or_else_async(fb))
        'bar'

        :param fallback: An async callable to generate a result if this in Error
        :return: Either the held value or the fallback value
        """
        raise NotImplementedError()

    def unwrap_err(self, msg: str | None = None) -> E:
        """Return the Error, or throw an UnwrapError

        :param msg: The message to add to the UnwrapError, otherwise a default is used
        :raises UnwrapError: Thrown if this is a Success
        :return: The held error
        """
        raise NotImplementedError()

    def map(self, cb: Callable[[T], U]) -> Result[U, E]:
        """Transform the held value or return the Error unchanged.

        :param cb: A callback taking the held success type and returning a new one
        :return: A result with the transformed Success or an unchanged Error
        """
        raise NotImplementedError()

    async def map_async(self, cb: Callable[[T], Awaitable[U]]) -> Result[U, E]:
        """Transform the held value or return the Error unchanged.

        >>> import asyncio
        >>> async def cb(v: int) -> str:
        ...     return str(v)
        >>> asyncio.run(Success(5).map_async(cb)).unwrap()
        '5'
        >>> asyncio.run(Error('ah!').map_async(cb)).unwrap_err()
        'ah!'

        :param cb: An async callback taking the held success type and returning
            a new one
        :return: A result with the transformed Success or an unchanged Error
        """
        raise NotImplementedError()

    def map_err(self, cb: Callable[[E], F]) -> Result[T, F]:
        """Transform the held error or return the success unchanged.

        :param cb: A callback taking the held error type and returning a new one
        :return: A result with the transformed Error or an unchanged Success
        """
        raise NotImplementedError()

    async def map_err_async(self, cb: Callable[[E], Awaitable[F]]) -> Result[T, F]:
        """Transform the held error or return the success unchanged,
        asynchronously.

        >>> import asyncio
        >>> async def cb(v: str) -> str:
        ...     return 'oh, no! ' + v
        >>> asyncio.run(Success(5).map_err_async(cb)).unwrap()
        5
        >>> asyncio.run(Error('ah!').map_err_async(cb)).unwrap_err()
        'oh, no! ah!'

        :param cb: An asynchronous callback taking the held error type and returning a new one
        :return: A result with the transformed Error or an unchanged Success
        """
        raise NotImplementedError()

    def map_or(self, default: U, cb: Callable[[T], U]) -> U:
        """Transform the held value or return the default.

        :param default: A value to use of Error
        :param cb: A callback to transform the held value of a Success
        :return: The fallback value or the transformed held value
        """
        raise NotImplementedError()

    async def map_or_async(self, default: U, cb: Callable[[T], Awaitable[U]]) -> U:
        """Transform the held value or return the default.

        >>> import asyncio
        >>> async def cb(v: int) -> int:
        ...     return v + 5
        >>> asyncio.run(Success(5).map_or_async(0, cb))
        10
        >>> asyncio.run(Error('ah!').map_or_async(0, cb))
        0

        :param default: A value to use of Error
        :param cb: An async callback to transform the held value of a Success
        :return: The fallback value or the transformed held value
        """
        raise NotImplementedError()

    def map_or_else(self, default: Callable[[], U], cb: Callable[[T], U]) -> U:
        """Transform the held value or return the calculated default

        :param default: A callable returning a value for Error
        :param cb: A callback to transform the held value of a Success
        :return: The fallback value or the transformed held value
        """
        raise NotImplementedError()

    async def map_or_else_async(
            self, default: Callable[[], Awaitable[U]], cb: Callable[[T], Awaitable[U]]) -> U:
        """Transform the held value or return the calculated default

        >>> import asyncio
        >>> async def cb(v: int) -> int:
        ...     return v + 5
        >>> async def fb() -> int:
        ...     return 0
        >>> asyncio.run(Success(5).map_or_else_async(fb, cb))
        10
        >>> asyncio.run(Error('ah!').map_or_else_async(fb, cb))
        0

        :param default: An async callable returning a value for Error
        :param cb: An async callback to transform the held value of a Success
        :return: The fallback value or the transformed held value
        """
        raise NotImplementedError()

    def and_then(self, cb: Callable[[T], Result[U, E]]) -> Result[U, E]:
        """Run the callback if this is a Success, otherwise return the Err unchanged

        :param cb: A callback run on the held value of a Success
        :return: a new Result with a transformed value or the error
        """
        raise NotImplementedError()

    async def and_then_async(self, cb: Callable[[T], Awaitable[Result[U, E]]]) -> Result[U, E]:
        """Run the callback if this is a Success, otherwise return the Err unchanged

        >>> import asyncio
        >>> async def cb(v: int) -> Result[int, E]:
        ...     return Success(v + 5)
        >>> asyncio.run(Success(5).and_then_async(cb)).unwrap()
        10
        >>> asyncio.run(Error('ah').and_then_async(cb)).unwrap_err()
        'ah'

        :param cb: An async callback run on the held value of a Success
        :return: a new Result with a transformed value or the error
        """
        raise NotImplementedError()

    def or_else(self, cb: Callable[[E], Result[T, F]]) -> Result[T, F]:
        """Run the callback if this is an Error, otherwise return the Success unchanged

        :param cb: A callback run on the held value of a Error
        :return: a new Result with a transformed value or the Success
        """
        raise NotImplementedError()

    async def or_else_async(self, cb: Callable[[E], Awaitable[Result[T, F]]]) -> Result[T, F]:
        """Run the callback if this is an Error, otherwise return the Success unchanged

        >>> import asyncio
        >>> async def cb(e: E) -> Result[int, E]:
        ...     return Success(0)
        >>> asyncio.run(Success(5).or_else_async(cb)).unwrap()
        5
        >>> asyncio.run(Error('ah').or_else_async(cb)).unwrap()
        0

        :param cb: An async callback run on the held value of a Error
        :return: a new Result with a transformed value or the Success
        """
        raise NotImplementedError()

    def err(self) -> Maybe[E]:
        """Transform an Result[T, E] into a Maybe[E]

        A Success will be mapped to Nothing[E], while and Error becomes Something[E]

        :return: A Maybe with the error
        """
        raise NotImplementedError()

    def ok(self) -> Maybe[T]:  # pylint: disable=invalid-name
        """Transform an Result[T, E] into a Maybe[T]

        A Success will be mapped to Something[T], while and Error becomes Nothing[T]

        :return: A Maybe with the held value
        """
        raise NotImplementedError()

    def propagate(self) -> T:
        """Get the value, or propagate an error up the stack.

        This is achieved by throwing a special Exception, which is caught using
        the :func:`stop` decorator. This is ugly and an abuse of Exceptions,
        (control flow through Exceptions, which is bad, m'kay?). However, this
        is the only obvious portable way to implement this in Python.

        :return: The held value of a Success

        >>> @stop
        ... def func() -> Result[bool, int]:
        ...     r: Result[str, int] = Error(5)
        ...     x = r.propagate()  # is now str, or thrown
        ...     return x == 'foo'
        ...
        >>> func()
        Error(5)
        """
        raise NotImplementedError()


@dataclass(slots=True, frozen=True)
class Error(Result[T, E]):

    """A Result holding an Error."""

    _held: E

    def __bool__(self) -> bool:
        return False

    def __repr__(self) -> str:
        return f'Error({self._held!r})'

    @staticmethod
    def is_ok() -> bool:
        return False

    @staticmethod
    def is_err() -> bool:
        return True

    def unwrap(self, msg: str | None = None) -> T:
        e: Exception
        if isinstance(self._held, Exception):
            e = self._held
        else:
            e = ErrorWrapper(self._held)
        raise UnwrapError(msg or 'Attempted to unwrap an Error') from e

    def unwrap_or(self, fallback: T) -> T:
        return fallback

    def unwrap_or_else(self, fallback: Callable[[], T]) -> T:
        return fallback()

    async def unwrap_or_else_async(self, fallback: Callable[[], Awaitable[T]]) -> T:
        return await fallback()

    def unwrap_err(self, msg: str | None = None) -> E:
        return self._held

    def map(self, cb: Callable[[T], U]) -> Result[U, E]:
        return Error(self._held)

    async def map_async(self, cb: Callable[[T], Awaitable[U]]) -> Result[U, E]:
        return Error(self._held)

    def map_err(self, cb: Callable[[E], F]) -> Result[T, F]:
        return Error(cb(self._held))

    async def map_err_async(self, cb: Callable[[E], Awaitable[F]]) -> Result[T, F]:
        return Error(await cb(self._held))

    def map_or(self, default: U, cb: Callable[[T], U]) -> U:
        return default

    async def map_or_async(self, default: U, cb: Callable[[T], Awaitable[U]]) -> U:
        return default

    def map_or_else(self, default: Callable[[], U], cb: Callable[[T], U]) -> U:
        return default()

    async def map_or_else_async(
            self, default: Callable[[], Awaitable[U]], cb: Callable[[T], Awaitable[U]]) -> U:
        return await default()

    def and_then(self, cb: Callable[[T], Result[U, E]]) -> Result[U, E]:
        return Error(self._held)

    async def and_then_async(self, cb: Callable[[T], Awaitable[Result[U, E]]]) -> Result[U, E]:
        return Error(self._held)

    def or_else(self, cb: Callable[[E], Result[T, F]]) -> Result[T, F]:
        return cb(self._held)

    async def or_else_async(self, cb: Callable[[E], Awaitable[Result[T, F]]]) -> Result[T, F]:
        return await cb(self._held)

    def err(self) -> Maybe[E]:
        from .maybe import Something  # pylint: disable=import-outside-toplevel
        return Something(self._held)

    def ok(self) -> Maybe[T]:
        from .maybe import Nothing  # pylint: disable=import-outside-toplevel
        return Nothing()

    def propagate(self) -> T:
        raise Propagation(self._held)


@dataclass(slots=True, frozen=True)
class Success(Result[T, E]):

    """A Result holding an value."""

    _held: T

    def __bool__(self) -> bool:
        return True

    def __repr__(self) -> str:
        return f'Success({self._held!r})'

    @staticmethod
    def is_ok() -> bool:
        return True

    @staticmethod
    def is_err() -> bool:
        return False

    def unwrap(self, msg: str | None = None) -> T:
        return self._held

    def unwrap_or(self, fallback: T) -> T:
        return self._held

    def unwrap_or_else(self, fallback: Callable[[], T]) -> T:
        return self._held

    async def unwrap_or_else_async(self, fallback: Callable[[], Awaitable[T]]) -> T:
        return self._held

    def unwrap_err(self, msg: str | None = None) -> E:
        if msg is None:
            msg = 'Attempted to unwrap the error from a Success'
        raise UnwrapError(msg)

    def map(self, cb: Callable[[T], U]) -> Result[U, E]:
        return Success(cb(self._held))

    async def map_async(self, cb: Callable[[T], Awaitable[U]]) -> Result[U, E]:
        return Success(await cb(self._held))

    def map_err(self, cb: Callable[[E], F]) -> Result[T, F]:
        return Success(self._held)

    async def map_err_async(self, cb: Callable[[E], Awaitable[F]]) -> Result[T, F]:
        return Success(self._held)

    def map_or(self, default: U, cb: Callable[[T], U]) -> U:
        return cb(self._held)

    async def map_or_async(self, default: U, cb: Callable[[T], Awaitable[U]]) -> U:
        return await cb(self._held)

    def map_or_else(self, default: Callable[[], U], cb: Callable[[T], U]) -> U:
        return cb(self._held)

    async def map_or_else_async(
            self, default: Callable[[], Awaitable[U]], cb: Callable[[T], Awaitable[U]]) -> U:
        return await cb(self._held)

    def and_then(self, cb: Callable[[T], Result[U, E]]) -> Result[U, E]:
        return cb(self._held)

    async def and_then_async(self, cb: Callable[[T], Awaitable[Result[U, E]]]) -> Result[U, E]:
        return await cb(self._held)

    def or_else(self, cb: Callable[[E], Result[T, F]]) -> Result[T, F]:
        return Success(self._held)

    async def or_else_async(self, cb: Callable[[E], Awaitable[Result[T, F]]]) -> Result[T, F]:
        return Success(self._held)

    def err(self) -> Maybe[E]:
        from .maybe import Nothing  # pylint: disable=import-outside-toplevel
        return Nothing()

    def ok(self) -> Maybe[T]:
        from .maybe import Something  # pylint: disable=import-outside-toplevel
        return Something(self._held)

    def propagate(self) -> T:
        return self._held


def wrap_result(catch: type[Exception] | tuple[type[Exception], ...] = Exception
                ) -> Callable[[Callable[P, R]], Callable[P, Result[R, Exception]]]:
    """Decorator for wrapping throwing functions to return a Result instead

    This is meant for simple cases only, if you wish to have more complex error
    handling than simply catching all exceptions and putting them in the Result
    you will need to handle that yourself.

    :param f: A callable to wrap
    :param catch: A Exeption or tuple of Exceptions to catch, defaults to Exception
    :return: A result of T or the caught Exception(s) as E
    """

    def wrapper(f: Callable[P, R]) -> Callable[P, Result[R, Exception]]:
        @wraps(f)
        def inner(*args: P.args, **kwargs: P.kwargs) -> Result[R, Exception]:
            try:
                return Success(f(*args, **kwargs))
            except catch as e:  # pylint: disable=broad-exception-caught
                return Error(e)

        return inner

    return wrapper


def wrap_result_async(catch: type[Exception] | tuple[type[Exception], ...] = Exception
                      ) -> Callable[[Callable[P, Awaitable[R]]],
                                     Callable[P, Awaitable[Result[R, Exception]]]]:
    """Decorator for wrapping throwing functions to return a Result instead

    This is meant for simple cases only, if you wish to have more complex error
    handling than simply catching all exceptions and putting them in the Result
    you will need to handle that yourself.

    :param f: A callable to wrap
    :param catch: A Exeption or tuple of Exceptions to catch, defaults to Exception
    :return: A result of T or the caught Exception(s) as E
    """

    def wrapper(f: Callable[P, Awaitable[R]]) -> Callable[P, Awaitable[Result[R, Exception]]]:
        @wraps(f)
        async def inner(*args: P.args, **kwargs: P.kwargs) -> Result[R, Exception]:
            try:
                return Success(await f(*args, **kwargs))
            except catch as e:  # pylint: disable=broad-exception-caught
                return Error(e)

        return inner

    return wrapper


def unwrap_result(f: Callable[P, Result[R, E]]) -> Callable[P, R]:
    """Decorator for unwrapping Result returning functions, return the result or
    throwing the Exception.

    This is meant for simple cases only, if you wish to have more complex error
    handling than simply catching all exceptions and putting them in the Result
    you will need to handle that yourself.

    :param f: A callable to unwrap
    :raises ErrorWrapper: if E is not an Exception type
    :raises E: any values of E that Exceptions
    :return: the valu eof a Success
    """

    @wraps(f)
    def inner(*args: P.args, **kwargs: P.kwargs) -> R:
        result = f(*args, **kwargs)
        if result.is_ok():
            return result.unwrap()
        err = result.unwrap_err()
        if isinstance(err, Exception):
            raise err
        raise ErrorWrapper(err)

    return inner


def unwrap_result_async(f: Callable[P, Awaitable[Result[R, E]]]) -> Callable[P, Awaitable[R]]:
    """Decorator for unwrapping Result returning functions, return the result or
    throwing the Exception.

    This is meant for simple cases only, if you wish to have more complex error
    handling than simply catching all exceptions and putting them in the Result
    you will need to handle that yourself.

    :param f: A callable to unwrap
    :raises ErrorWrapper: if E is not an Exception type
    :raises E: any values of E that Exceptions
    :return: the value of a Success
    """

    @wraps(f)
    async def inner(*args: P.args, **kwargs: P.kwargs) -> R:
        result = await f(*args, **kwargs)
        if result.is_ok():
            return result.unwrap()
        err = result.unwrap_err()
        if isinstance(err, Exception):
            raise err
        raise ErrorWrapper(err)

    return inner


def stop(f: Callable[P, R]) -> Callable[P, Result[R, E]]:
    """Decorator for functions that use :meth:`Result.propagate`.

    This is required to catch the propagated Error, and ensure that it is
    returned instead of continuing to go up the stack.

    :param f: The function to wrap
    :return: The original function wrapped to handle Propagation Exceptions
    """

    @wraps(f)
    def inner(*args: P.args, **kwargs: P.kwargs) -> Result[R, E]:
        try:
            return Success(f(*args, **kwargs))
        except Propagation as e:
            return Error(e.err)

    return inner


def stop_async(f: Callable[P, Awaitable[R]]) -> Callable[P, Awaitable[Result[R, E]]]:
    """Decorator for async functions that use :meth:`Result.propagate`.

    This is required to catch the propagated Error, and ensure that it is
    returned instead of continuing to go up the stack.

    >>> async def g() -> Result[str, str]:
    ...     return Error('err!')

    >>> @stop_async
    ... async def f() -> Result[int, str]:
    ...     v = await g()
    ...     v.propagate()
    ...     x = v.map(int)
    ...     return x.map(lambda x: x + 10).map_err(lambda e: 'got: ' + e)

    >>> import asyncio
    >>> asyncio.run(f())
    Error('err!')

    :param f: The async function to wrap
    :return: The original function wrapped to handle Propagation Exceptions
    """

    @wraps(f)
    async def inner(*args: P.args, **kwargs: P.kwargs) -> Result[R, E]:
        try:
            return Success(await f(*args, **kwargs))
        except Propagation as e:
            return Error(e.err)

    return inner



# SPDX-License-Identifier: MIT
# Copyright © 2023-2024 Dylan Baker

"""An implementation of an Option type."""

from __future__ import annotations
from dataclasses import dataclass
from functools import wraps
from typing import TYPE_CHECKING, TypeVar, ParamSpec, Generic

if TYPE_CHECKING:
    from typing import Callable, Awaitable

    from .result import Result

P = ParamSpec('P')
R = TypeVar('R')
E = TypeVar('E')
T = TypeVar('T')
U = TypeVar('U')

__all__ = [
    'EmptyMaybeError',
    'Maybe',
    'Nothing',
    'Something',
    'maybe',
    'unwrap_maybe',
    'unwrap_maybe_async',
    'wrap_maybe',
    'wrap_maybe_async',
]


class EmptyMaybeError(Exception):

    """Raised when calling :meth:`Maybe.unwrap` on an empty Maybe.

    It is advised to *not* catch this exception, as it is usually the result of
    an incorrect assumption in the code, or it truly is fatal.

    In the former case replacing :meth:`Maybe.unwrap` with
    :meth:`Maybe.unwrap_or` or :meth:`Maybe.get` may be mor appropriate.
    """


class Maybe(Generic[T]):

    """Base Class for Option, do not directly instantiate.

    This class is not useful for instantiation, instead use either
    :class:`Something` or :class:`Nothing`. This class is useful for isinstance
    """

    @staticmethod
    def is_something() -> bool:
        """Is this Something?

        It is *strongly* recommended to use this method an not
        :func:`isinstance`.

        >>> maybe('foo').is_something()
        True

        >>> maybe(None).is_something()
        False

        :return: True if this is Something otherwise False
        """
        raise NotImplementedError()

    @staticmethod
    def is_nothing() -> bool:
        """Is this Nothing?

        It is *strongly* recommended to use this method an not
        :func:`isinstance`.

        >>> maybe('foo').is_nothing()
        False

        >>> maybe(None).is_nothing()
        True

        :return: True if this is Nothing otherwise False
        """
        raise NotImplementedError()

    def map(self, cb: Callable[[T], U]) -> Maybe[U]:
        """Transforms the held value using the callback

        If this Maybe is Nothing, then Nothing is returned

        >>> maybe(500).map(str).unwrap()
        '500'

        >>> maybe(None).map(str).is_nothing()
        True

        :param cb: A callback transforming the held value from T to U
        :return: A new Maybe holding the transformed value
        """
        raise NotImplementedError()

    async def map_async(self, cb: Callable[[T], Awaitable[U]]) -> Maybe[U]:
        """Transforms the held value using the callback asynchronously.

        If this Maybe is Nothing, then Nothing is returned

        >>> import asyncio
        >>> async def foo(a: int) -> str:
        ...     return str(a)
        >>> asyncio.run(maybe(500).map_async(foo)).unwrap()
        '500'
        >>> asyncio.run(maybe(None).map_async(foo)).is_nothing()
        True

        :param cb: A callback transforming the held value from T to awaitable U
        :return: A new Maybe holding the transformed value
        """
        raise NotImplementedError()


    def map_or(self, cb: Callable[[T], U], fallback: U) -> Maybe[U]:
        """Transform the held value using the callback, or use the fallback
        value.

        >>> maybe(500).map_or(str, '0').unwrap()
        '500'

        >>> maybe(None).map_or(str, '0').unwrap()
        '0'

        :param cb: A callback which will transform Something[T] into Something[U]
        :param fallback: A value to use for Nothing
        :return: A Something containing the transformation or the fallback value
        """
        raise NotImplementedError()

    async def map_or_async(self, cb: Callable[[T], Awaitable[U]], fallback: U) -> Maybe[U]:
        """Transform the held value using the callback, or use the fallback
        value asynchronously.

        >>> import asyncio
        >>> async def foo(a: int) -> str:
        ...     return str(a)

        >>> asyncio.run(maybe(500).map_or_async(foo, '0')).unwrap()
        '500'

        >>> asyncio.run(maybe(None).map_or_async(foo, '0')).unwrap()
        '0'

        :param cb: A callback which will transform Something[T] into Something[U]
        :param fallback: A value to use for Nothing
        :return: A Something containing the transformation or the fallback value
        """
        raise NotImplementedError()

    def map_or_else(self, cb: Callable[[T], U], fallback: Callable[[], U]) -> Maybe[U]:
        """Transform the held value using the callback, or use the fallback
        value.

        >>> maybe(500).map_or_else(str, lambda: '0').unwrap()
        '500'

        >>> maybe(None).map_or_else(str, lambda: '0').unwrap()
        '0'

        :param cb: A callback which will transform Something[T] into Something[U]
        :param fallback: callable returning a value U
        :return: A Something containing the transformation or the fallback value
        """
        raise NotImplementedError()

    async def map_or_else_async(
            self, cb: Callable[[T], Awaitable[U]], fallback: Callable[[], Awaitable[U]]) -> Maybe[U]:  # pylint: disable=line-too-long
        """Transform the held value using the callback, or use the fallback
        value asynchronously.

        >>> import asyncio
        >>> async def fb() -> str:
        ...     return '0'

        >>> async def cb(v: int) -> str:
        ...     return str(v)

        >>> asyncio.run(maybe(500).map_or_else_async(cb, fb)).unwrap()
        '500'

        >>> asyncio.run(maybe(None).map_or_else_async(cb, fb)).unwrap()
        '0'

        :param cb: A callback which will transform Something[T] into Something[U]
        :param fallback: callable returning a value U
        :return: A Something containing the transformation or the fallback value
        """
        raise NotImplementedError()

    def get(self, fallback: T | None = None) -> T | None:
        """Get the held value.

        Works much like Python's normal `.get()` methods, but never throws.

        >>> maybe('500').get()
        '500'

        >>> maybe(None).get('a')
        'a'

        :param fallback: A value to use if this is Nothing
        :return: The value or fallback
        """
        raise NotImplementedError()

    def unwrap(self, msg: str | None = None) -> T:
        """Get the held value or throw an Exception.

        >>> maybe('foo').unwrap()
        'foo'

        >>> maybe(None).unwrap()
        Traceback (most recent call last):
          ...
        simple_monads.maybe.EmptyMaybeError: Attempted to unwrap Nothing

        >>> maybe(None).unwrap("Expected a result")
        Traceback (most recent call last):
          ...
        simple_monads.maybe.EmptyMaybeError: Expected a result

        :param msg: The error message, otherwise a default is used
        :raises EmptyMaybeError: If this is Nothing
        :return: The held value
        """
        raise NotImplementedError()

    def unwrap_or(self, fallback: T) -> T:
        """Get the value or a fallback value.

        Unlike get() doesn't provide a fallback of None, which narrows type
        checking.

        >>> maybe(500).unwrap_or(0)
        500

        >>> maybe(None).unwrap_or(0)
        0

        :param fallback: The fallback to return
        :return: The held value or the fallback
        """
        raise NotImplementedError()

    def unwrap_or_else(self, fallback: Callable[[], T]) -> T:
        """Get the value or call the fallback to get a value

        >>> maybe(500).unwrap_or_else(lambda: 0)
        500

        >>> maybe(None).unwrap_or_else(lambda: 0)
        0

        :param fallback: A callable returning a type T
        :return: The held value or the fallback
        """
        raise NotImplementedError()

    async def unwrap_or_else_async(self, fallback: Callable[[], Awaitable[T]]) -> T:
        """Get the value or call the fallback to get a value

        >>> import asyncio
        >>> async def fb():
        ...     return 0

        >>> asyncio.run(maybe(500).unwrap_or_else_async(fb))
        500

        >>> asyncio.run(maybe(None).unwrap_or_else_async(fb))
        0

        :param fallback: An async callable returning a type T
        :return: The held value or the fallback
        """
        raise NotImplementedError()

    def and_then(self, cb: Callable[[T], Maybe[U]]) -> Maybe[U]:
        """Run a callback on the value if it is Something

        >>> maybe(500).and_then(lambda x: maybe(0)).unwrap()
        0

        :param cb: A callback to run on the held value or Something()
        :return: A Maybe[U] with the result of the callback or nothing
        """
        raise NotImplementedError()

    async def and_then_async(self, cb: Callable[[T], Awaitable[Maybe[U]]]) -> Maybe[U]:
        """Run a callback on the value if it is Something

        >>> import asyncio
        >>> async def cb(v: int):
        ...     return Something(str(v))
        >>> asyncio.run(maybe(500).and_then_async(cb)).unwrap()
        '500'

        :param cb: A callback to run on the held value or Something()
        :return: A Maybe[U] with the result of the callback or nothing
        """
        raise NotImplementedError()

    def or_else(self, fallback: Callable[[], Maybe[T]]) -> Maybe[T]:
        """Run a callback to get a value if this is Nothing or return self.

        >>> maybe(500).or_else(lambda: maybe(100)).unwrap()
        500

        >>> maybe(None).or_else(lambda: maybe(100)).unwrap()
        100

        :param fallback: A callback to run if this is Nothing returning a
            Maybe[T]
        :return: A Maybe[T], which is self unchanged if this Something,
            otherwise the result of fallback
        """
        raise NotImplementedError()

    async def or_else_async(self, fallback: Callable[[], Awaitable[Maybe[T]]]) -> Maybe[T]:
        """Run a callback to get a value if this is Nothing or return self.

        >>> import asyncio
        >>> async def cb():
        ...     return maybe(100)
        >>> asyncio.run(maybe(500).or_else_async(cb)).unwrap()
        500

        >>> asyncio.run(maybe(None).or_else_async(cb)).unwrap()
        100

        :param fallback: An async callback to run if this is Nothing returning a
            Maybe[T]
        :return: A Maybe[T], which is self unchanged if this Something,
            otherwise the result of fallback
        """
        raise NotImplementedError()

    def ok_or(self, err: E) -> Result[T, E]:
        """Convert this Option to a Result.

        If the Option is Something, that will be placed in the Success value,
        otherwise the Error value of E will be used.

        >>> maybe(0).ok_or("WHAT!").unwrap()
        0

        >>> maybe(None).ok_or("WHAT!").unwrap_err()
        'WHAT!'

        :param err: An error if this is Nothing
        :return: A result with the held value as a Success or an Error
        """
        raise NotImplementedError()

    def ok_or_else(self, err: Callable[[], E]) -> Result[T, E]:
        """Convert this Option to a Result.

        If the Option is Something, that will be placed in the Success value,
        otherwise the Error value of E will be used.

        >>> maybe(0).ok_or_else(lambda: "WHAT!").unwrap()
        0

        >>> maybe(None).ok_or_else(lambda: "WHAT!").unwrap_err()
        'WHAT!'

        :param err: An callable returning a type E
        :return: A result with the held value as a Success or an Error
        """
        raise NotImplementedError()

    async def ok_or_else_async(self, err: Callable[[], Awaitable[E]]) -> Result[T, E]:
        """Convert this Option to a Result asynchronously.

        If the Option is Something, that will be placed in the Success value,
        otherwise the Error value of E will be used.

        >>> import asyncio
        >>> async def cb():
        ...     return "WHAT!"
        >>> asyncio.run(maybe(0).ok_or_else_async(cb)).unwrap()
        0

        >>> asyncio.run(maybe(None).ok_or_else_async(cb)).unwrap_err()
        'WHAT!'

        :param err: An async callable returning a type E
        :return: A result with the held value as a Success or an Error
        """
        raise NotImplementedError()



@dataclass(slots=True, frozen=True)
class Something(Maybe[T]):

    """A Maybe that holds a value."""

    _held: T

    def __bool__(self) -> bool:
        return True

    def __repr__(self) -> str:
        return f"Something({self._held!r})"

    @staticmethod
    def is_something() -> bool:
        return True

    @staticmethod
    def is_nothing() -> bool:
        return False

    def map(self, cb: Callable[[T], U]) -> Maybe[U]:
        return Something(cb(self._held))

    async def map_async(self, cb: Callable[[T], Awaitable[U]]) -> Maybe[U]:
        return Something(await cb(self._held))

    def map_or(self, cb: Callable[[T], U], fallback: U) -> Maybe[U]:
        return Something(cb(self._held))

    async def map_or_async(self, cb: Callable[[T], Awaitable[U]], fallback: U) -> Maybe[U]:
        return Something(await cb(self._held))

    def map_or_else(self, cb: Callable[[T], U], fallback: Callable[[], U]) -> Maybe[U]:
        return Something(cb(self._held))

    async def map_or_else_async(
            self, cb: Callable[[T], Awaitable[U]], fallback: Callable[[], Awaitable[U]]) -> Maybe[U]:  # pylint: disable=line-too-long
        return Something(await cb(self._held))

    def get(self, fallback: T | None = None) -> T | None:
        return self._held

    def unwrap(self, msg: str | None = None) -> T:
        return self._held

    def unwrap_or(self, fallback: T) -> T:
        return self._held

    def unwrap_or_else(self, fallback: Callable[[], T]) -> T:
        return self._held

    async def unwrap_or_else_async(self, fallback: Callable[[], Awaitable[T]]) -> T:
        return self._held

    def and_then(self, cb: Callable[[T], Maybe[U]]) -> Maybe[U]:
        return cb(self._held)

    async def and_then_async(self, cb: Callable[[T], Awaitable[Maybe[U]]]) -> Maybe[U]:
        return await cb(self._held)

    def or_else(self, fallback: Callable[[], Maybe[T]]) -> Maybe[T]:
        return self

    async def or_else_async(self, fallback: Callable[[], Awaitable[Maybe[T]]]) -> Maybe[T]:
        return self

    def ok_or(self, err: E) -> Result[T, E]:
        from .result import Success  # pylint: disable=import-outside-toplevel
        return Success(self._held)

    def ok_or_else(self, err: Callable[[], E]) -> Result[T, E]:
        from .result import Success  # pylint: disable=import-outside-toplevel
        return Success(self._held)

    async def ok_or_else_async(self, err: Callable[[], Awaitable[E]]) -> Result[T, E]:
        from .result import Success  # pylint: disable=import-outside-toplevel
        return Success(self._held)


@dataclass(slots=True, frozen=True)
class Nothing(Maybe[T]):

    """A Maybe that does not hold a value."""

    def __bool__(self) -> bool:
        return False

    @staticmethod
    def is_something() -> bool:
        return False

    @staticmethod
    def is_nothing() -> bool:
        return True

    def map(self, cb: Callable[[T], U]) -> Maybe[U]:
        return Nothing()

    async def map_async(self, cb: Callable[[T], Awaitable[U]]) -> Maybe[U]:
        return Nothing()

    def map_or(self, cb: Callable[[T], U], fallback: U) -> Maybe[U]:
        return Something(fallback)

    async def map_or_async(self, cb: Callable[[T], Awaitable[U]], fallback: U) -> Maybe[U]:
        return Something(fallback)

    def map_or_else(self, cb: Callable[[T], U], fallback: Callable[[], U]) -> Maybe[U]:
        return Something(fallback())

    async def map_or_else_async(
            self, cb: Callable[[T], Awaitable[U]], fallback: Callable[[], Awaitable[U]]) -> Maybe[U]:  # pylint: disable=line-too-long
        return Something(await fallback())

    def get(self, fallback: T | None = None) -> T | None:
        return fallback

    def unwrap(self, msg: str | None = None) -> T:
        if msg is None:
            msg = 'Attempted to unwrap Nothing'
        raise EmptyMaybeError(msg)

    def unwrap_or(self, fallback: T) -> T:
        return fallback

    def unwrap_or_else(self, fallback: Callable[[], T]) -> T:
        return fallback()

    async def unwrap_or_else_async(self, fallback: Callable[[], Awaitable[T]]) -> T:
        return await fallback()

    def and_then(self, cb: Callable[[T], Maybe[U]]) -> Maybe[U]:
        return Nothing()

    async def and_then_async(self, cb: Callable[[T], Awaitable[Maybe[U]]]) -> Maybe[U]:
        return Nothing()

    def or_else(self, fallback: Callable[[], Maybe[T]]) -> Maybe[T]:
        return fallback()

    async def or_else_async(self, fallback: Callable[[], Awaitable[Maybe[T]]]) -> Maybe[T]:
        return await fallback()

    def ok_or(self, err: E) -> Result[T, E]:
        from .result import Error  # pylint: disable=import-outside-toplevel
        return Error(err)

    def ok_or_else(self, err: Callable[[], E]) -> Result[T, E]:
        from .result import Error  # pylint: disable=import-outside-toplevel
        return Error(err())

    async def ok_or_else_async(self, err: Callable[[], Awaitable[E]]) -> Result[T, E]:
        from .result import Error  # pylint: disable=import-outside-toplevel
        return Error(await err())


def maybe(result: T | None) -> Maybe[T]:
    """Convenience function to convert T | None into Maybe[T].

    This can convert python code using the standard T | None Optional.
    This works correctly only when None is not a valid member of T

    >>> maybe(0)
    Something(0)

    >>> maybe(None)
    Nothing()

    :param result: A None or T type to wrap
    :return: Nothing if result is None, else Something[T](result)
    """
    if result is None:
        return Nothing()
    return Something(result)


def wrap_maybe_async(f: Callable[P, Awaitable[R | None]]) -> Callable[P, Awaitable[Maybe[R]]]:
    """Decorator (or wrapper) for asynchronous python code.

    Converts code returning Awaitable[T | None] to return Awaitable[Maybe[T]]

    >>> import asyncio

    >>> @wrap_maybe_async
    ... async def func(arg: bool) -> str | None:
    ...     if arg:
    ...         return "yes!"
    ...     return None
    ...
    >>> asyncio.run(func(True))
    Something('yes!')

    >>> asyncio.run(func(False))
    Nothing()

    :param f: An async callable returning a type R or None
    :return: A new async callable return :class:`Maybe[R]`, where a non-null are
        :class:`Success`, and None is :class:`Nothing`
    """

    @wraps(f)
    async def inner(*args: P.args, **kwargs: P.kwargs) -> Maybe[R]:
        return maybe(await f(*args, **kwargs))

    return inner


def wrap_maybe(f: Callable[P, R | None]) -> Callable[P, Maybe[R]]:
    """Decorator (or wrapper) for common python code.

    Converts code returning T | None to return Maybe[T]

    >>> @wrap_maybe
    ... def func(arg: bool) -> str | None:
    ...     if arg:
    ...         return "yes!"
    ...     return None
    ...
    >>> func(True)
    Something('yes!')

    >>> func(False)
    Nothing()

    >>> import os
    >>> f = wrap_maybe(os.environ.get)
    >>> f("Totally not there")
    Nothing()

    :param f: A callable returning a type R or None
    :return: A new callable return :class:`Maybe[R]`, where a non-null are
        :class:`Success`, and None is :class:`Nothing`
    """

    @wraps(f)
    def inner(*args: P.args, **kwargs: P.kwargs) -> Maybe[R]:
        return maybe(f(*args, **kwargs))

    return inner


def unwrap_maybe_async(f: Callable[P, Awaitable[Maybe[R]]]) -> Callable[P, Awaitable[R | None]]:
    """Decorator (or wrapper) to convert back to common asynchronous Python.

    Converts code returning Awaitable[Maybe[T]] to Awaitable[T | None].

    This is meant to ease transitioning a codebase to using simple_monads, but
    allowing code to internally use Maybe, but return common Python Optional

    >>> import asyncio
    >>> async def raw(arg: str) -> Maybe[str]:
    ...     if arg:
    ...         return Something(arg)
    ...     return Nothing()

    >>> f = unwrap_maybe_async(raw)
    >>> asyncio.run(f("foo"))
    'foo'

    >>> @unwrap_maybe_async
    ... async def g(arg: str) -> str | None:
    ...     return await raw(arg)

    >>> asyncio.run(g("foo"))
    'foo'

    :param f: An async callable returning a :class:`Maybe[T]`
    :return: A new async callable returning a `T | None`
    """

    @wraps(f)
    async def inner(*args: P.args, **kwargs: P.kwargs) -> R | None:
        return (await f(*args, **kwargs)).get()

    return inner


def unwrap_maybe(f: Callable[P, Maybe[R]]) -> Callable[P, R | None]:

    """Decorator (or wrapper) to convert back to common Python.

    Converts code returning Maybe[T] to T | None.

    This is meant to ease transitioning a codebase to using simple_monads, but
    allowing code to internally use Maybe, but return common Python Optional

    >>> def raw(arg: str) -> Maybe[str]:
    ...     if arg:
    ...         return Something(arg)
    ...     return Nothing()

    >>> f = unwrap_maybe(raw)
    >>> f("foo")
    'foo'

    >>> @unwrap_maybe
    ... def g(arg: str) -> str | None:
    ...     return raw(arg)

    >>> g("foo")
    'foo'

    :param f: A callable returning a :class:`Maybe[T]`
    :return: A new callable returning a `T | None`
    """

    @wraps(f)
    def inner(*args: P.args, **kwargs: P.kwargs) -> R | None:
        return f(*args, **kwargs).get()

    return inner



# SPDX-License-Identifier: MIT
# Copyright © 2023-2024 Dylan Baker

from __future__ import annotations
from typing import cast

import pytest

from simple_monads.result import *

class TestResult:

    class TestUnwrap:

        def test_error(self) -> None:
            with pytest.raises(UnwrapError, match='Attempted to unwrap an Error'):
                Error(Exception('foo')).unwrap()

        def test_error_msg(self) -> None:
            msg = 'test message'
            with pytest.raises(UnwrapError, match=msg):
                Error(Exception('foo')).unwrap(msg)

        def test_success(self) -> None:
            assert Success('foo').unwrap() == 'foo'

    class TestUnwrapOr:

        def test_error(self) -> None:
            e: Result[str, Exception] = Error(Exception('foo'))
            assert e.unwrap_or('bar') == 'bar'

        def test_success(self) -> None:
            assert Success('foo').unwrap_or('bar') == 'foo'

    class TestUnwrapOrElse:

        def test_error(self) -> None:
            e: Result[str, Exception] = Error(Exception('foo'))
            assert e.unwrap_or_else(lambda: 'bar') == 'bar'

        def test_success(self) -> None:
            assert Success('foo').unwrap_or_else(lambda: 'bar') == 'foo'

    class TestUnwrapOrElseAsync:

        @pytest.mark.asyncio
        async def test_error(self) -> None:
            async def fb() -> str:
                return 'bar'

            e: Result[str, Exception] = Error(Exception('foo'))
            assert await e.unwrap_or_else_async(fb) == 'bar'

        @pytest.mark.asyncio
        async def test_success(self) -> None:
            async def fb() -> str:
                return 'bar'
            assert await Success('foo').unwrap_or_else_async(fb) == 'foo'

    class TestBool:

        def test_error(self) -> None:
            e: Result[str, Exception] = Error(Exception('foo'))
            assert not e

        def test_success(self) -> None:
            assert Success('foo')

    class TestIsErr:

        def test_error(self) -> None:
            e: Result[str, Exception] = Error(Exception('foo'))
            assert e.is_err()

        def test_success(self) -> None:
            assert not Success('foo').is_err()

    class TestIsOk:

        def test_error(self) -> None:
            e: Result[str, Exception] = Error(Exception('foo'))
            assert not e.is_ok()

        def test_success(self) -> None:
            assert Success('foo').is_ok()

    class TestUnwrapErr:

        def test_error(self) -> None:
            err = Exception('foo')
            e: Result[str, Exception] = Error(err)
            assert e.unwrap_err() is err

        def test_success(self) -> None:
            with pytest.raises(UnwrapError, match='Attempted to unwrap the error from a Success'):
                Success('foo').unwrap_err()

        def test_success_with_msg(self) -> None:
            msg = 'test message'
            with pytest.raises(UnwrapError, match=msg):
                Success('foo').unwrap_err(msg)

    class TestMap:

        def test_error(self) -> None:
            e: Result[str, Exception] = Error(Exception('foo'))
            assert e.map(int) == cast(Result[int, Exception], e)

        def test_success(self) -> None:
            assert Success('1').map(int) == Success(1)

    class TestMapAsync:

        @pytest.mark.asyncio
        async def test_error(self) -> None:
            async def cb(v: str) -> int:
                pytest.fail()

            e: Result[str, Exception] = Error(Exception('foo'))
            assert await e.map_async(cb) == cast(Result[int, Exception], e)

        @pytest.mark.asyncio
        async def test_success(self) -> None:
            async def cb(v: str) -> int:
                return int(v)

            assert await Success('1').map_async(cb) == Success(1)

    class TestMapErr:

        def test_error(self) -> None:
            e: Result[str, str] = Error('1')
            assert e.map_err(int) == Error(1)

        def test_success(self) -> None:
            s: Success[str, str] = Success('1')
            assert s.map_err(int) == cast(Success[str, int], s)

    class TestMapErrAsync:

        @pytest.mark.asyncio
        async def test_error(self) -> None:
            async def cb(v: str) -> int:
                return int(v)
            e: Result[str, str] = Error('1')
            assert await e.map_err_async(cb) == Error(1)

        @pytest.mark.asyncio
        async def test_success(self) -> None:
            async def cb(v: str) -> int:
                pytest.fail()
            s: Success[str, str] = Success('1')
            assert await s.map_err_async(cb) == cast(Success[str, int], s)

    class TestMapOr:

        def test_error(self) -> None:
            e: Result[str, str] = Error('1')
            assert e.map_or('foo', lambda _: 'bar') == 'foo'

        def test_success(self) -> None:
            s: Success[str, str] = Success('1')
            assert s.map_or('foo', lambda _: 'bar') == 'bar'

    class TestMapOrAsync:

        @pytest.mark.asyncio
        async def test_error(self) -> None:
            async def cb(v: str) -> str:
                pytest.fail()

            e: Result[str, str] = Error('1')
            assert await e.map_or_async('foo', cb) == 'foo'

        @pytest.mark.asyncio
        async def test_success(self) -> None:
            async def cb(v: str) -> str:
                return 'bar'
            s: Success[str, str] = Success('1')
            assert await s.map_or_async('foo', cb) == 'bar'

    class TestMapOrElse:

        def test_error(self) -> None:
            e: Result[str, str] = Error('1')
            assert e.map_or_else(lambda: 'foo', lambda _: 'bar') == 'foo'

        def test_success(self) -> None:
            s: Success[str, str] = Success('1')
            assert s.map_or_else(lambda: 'foo', lambda _: 'bar') == 'bar'

    class TestMapOrElseASync:

        @staticmethod
        async def _cb(v: int) -> int:
            return v + 5

        @staticmethod
        async def _fb() -> int:
            return 3

        @pytest.mark.asyncio
        async def test_error(self) -> None:
            e: Result[int, int] = Error(1)
            assert await e.map_or_else_async(self._fb, self._cb) == 3

        @pytest.mark.asyncio
        async def test_success(self) -> None:
            s: Success[int, str] = Success(1)
            assert await s.map_or_else_async(self._fb, self._cb) == 6

    class TestAndThen:

        @staticmethod
        def _cb(res: str) -> Result[int, str]:
            return Success(int(res))

        def test_error(self) -> None:
            e: Result[str, str] = Error('1')
            assert e.and_then(self._cb) == cast(Result[int, str], e)

        def test_success(self) -> None:
            s: Success[str, str] = Success('1')
            assert s.and_then(self._cb) == Success(1)

    class TestAndThenAsync:

        @staticmethod
        async def _cb(res: str) -> Result[int, str]:
            return Success(int(res))

        @pytest.mark.asyncio
        async def test_error(self) -> None:
            e: Result[str, str] = Error('1')
            assert await e.and_then_async(self._cb) == cast(Result[int, str], e)

        @pytest.mark.asyncio
        async def test_success(self) -> None:
            s: Success[str, str] = Success('1')
            assert await s.and_then_async(self._cb) == Success(1)

    class TestOrElse:

        @staticmethod
        def _cb(res: str) -> Result[str, int]:
            return Error(int(res))

        def test_error(self) -> None:
            e: Result[str, str] = Error('1')
            assert e.or_else(self._cb) == Error(1)

        def test_success(self) -> None:
            s: Success[str, str] = Success('1')
            assert s.or_else(self._cb) == cast(Result[str, int], s)

    class TestOrElseAsync:

        @staticmethod
        async def _cb(res: str) -> Result[str, int]:
            return Error(int(res))

        @pytest.mark.asyncio
        async def test_error(self) -> None:
            e: Result[str, str] = Error('1')
            assert await e.or_else_async(self._cb) == Error(1)

        @pytest.mark.asyncio
        async def test_success(self) -> None:
            s: Success[str, str] = Success('1')
            assert await s.or_else_async(self._cb) == cast(Result[str, int], s)

    class TestErr:

        def test_error(self) -> None:
            e: Result[str, int] = Error(4)
            assert e.err().unwrap() == 4

        def test_success(self) -> None:
            s: Result[int, int] = Success(4)
            assert s.err().is_nothing()

    class TestOk:

        def test_error(self) -> None:
            e: Result[str, int] = Error(4)
            assert e.ok().is_nothing()

        def test_success(self) -> None:
            s: Result[int, int] = Success(4)
            assert s.ok().unwrap() == 4

    class TestMatch:

        def test_error(self) -> None:
            e: Result[str, int] = Error(4)
            match e:
                case Success(4):
                    pytest.fail()
                case Error(4):
                    assert True
                case _:
                    pytest.fail()

        def test_success(self) -> None:
            s: Result[int, int] = Success(4)
            match s:
                case Error(4):
                    pytest.fail()
                case Success(4):
                    assert True
                case _:
                    pytest.fail()


class TestWrapResult:

    def test_success(self) -> None:
        @wrap_result()
        def foo() -> str:
            return ''

        assert foo() == Success('')

    def test_error(self) -> None:
        @wrap_result(Exception)
        def foo() -> str:
            raise Exception('foo')

        assert isinstance(foo().unwrap_err(), Exception)

    def test_error_multi(self) -> None:
        @wrap_result((ValueError, RecursionError))
        def foo() -> str:
            raise ValueError('foo')

        assert isinstance(foo().unwrap_err(), ValueError)

    def test_error_uncaught(self) -> None:
        @wrap_result(ArithmeticError)
        def foo() -> str:
            raise ValueError('foo')

        with pytest.raises(ValueError, match='foo'):
            foo()


class TestWrapResultASync:

    @pytest.mark.asyncio
    async def test_success(self) -> None:
        @wrap_result_async()
        async def foo() -> str:
            return ''

        assert await foo() == Success('')

    @pytest.mark.asyncio
    async def test_error(self) -> None:
        @wrap_result_async(Exception)
        async def foo() -> str:
            raise Exception('foo')

        err = await foo()
        assert isinstance(err.unwrap_err(), Exception)

    @pytest.mark.asyncio
    async def test_error_multi(self) -> None:
        @wrap_result_async((ValueError, RecursionError))
        async def foo() -> str:
            raise ValueError('foo')

        err = await foo()
        assert isinstance(err.unwrap_err(), ValueError)

    @pytest.mark.asyncio
    async def test_error_uncaught(self) -> None:
        @wrap_result_async(ArithmeticError)
        async def foo() -> str:
            raise ValueError('foo')

        with pytest.raises(ValueError, match='foo'):
            await foo()


class TestUnwrapResult:

    def test_success(self) -> None:
        @unwrap_result
        def foo() -> Result[str, ValueError]:
            return Success('foo')

        assert foo() == 'foo'

    def test_error(self) -> None:
        @unwrap_result
        def foo() -> Result[str, ValueError]:
            return Error(ValueError('foo'))

        with pytest.raises(ValueError, match='foo'):
            foo()


class TestUnwrapResultAsync:

    @pytest.mark.asyncio
    async def test_success(self) -> None:
        @unwrap_result_async
        async def foo() -> Result[str, ValueError]:
            return Success('foo')

        assert await foo() == 'foo'

    @pytest.mark.asyncio
    async def test_error(self) -> None:
        @unwrap_result_async
        async def foo() -> Result[str, ValueError]:
            return Error(ValueError('foo'))

        with pytest.raises(ValueError, match='foo'):
            await foo()


class TestPpropgate:

    def test_prop(self) -> None:

        @stop
        def inner() -> str:
            r: Result[str, int] = Error(4)
            x = r.propagate()
            return x + 'bar'

        assert inner() == Error(4)

    def test_no_prop(self) -> None:

        @stop
        def inner() -> str:
            r: Result[str, int] = Success('foo')
            x = r.propagate()
            return x + 'bar'

        assert inner() == Success('foobar')


class TestPropogateAsync:

    @pytest.mark.asyncio
    async def test_prop(self) -> None:
        @stop_async
        async def inner() -> str:
            r: Result[str, int] = Error(4)
            x = r.propagate()
            return x + 'bar'

        assert await inner() == Error(4)

    @pytest.mark.asyncio
    async def test_no_prop(self) -> None:
        @stop_async
        async def inner() -> str:
            r: Result[str, int] = Success('foo')
            x = r.propagate()
            return x + 'bar'

        assert await inner() == Success('foobar')



# SPDX-License-Identifier: MIT
# Copyright © 2023-2024 Dylan Baker

from __future__ import annotations
from typing import TYPE_CHECKING, cast

import pytest

from simple_monads.maybe import *

if TYPE_CHECKING:
    from typing import Awaitable, Callable


class TestMaybe:

    class TestMap:

        def test_something(self) -> None:
            s = Something(1).map(str)
            assert s == Something('1')

        def test_nothing(self) -> None:
            s = Nothing().map(str)
            assert s == Nothing()

    class TestMapAsync:

        @staticmethod
        async def _helper(v: int) -> str:
            return str(v)

        @pytest.mark.asyncio
        async def test_something(self) -> None:
            s = await Something(1).map_async(self._helper)
            assert s == Something('1')

        @pytest.mark.asyncio
        async def test_nothing(self) -> None:
            s = await Nothing().map_async(self._helper)
            assert s == Nothing()

    class TestMapOr:

        def test_something(self) -> None:
            assert Something(1).map_or(str, '2') == Something('1')

        def test_nothing(self) -> None:
            assert Nothing().map_or(str, '2') == Something('2')

    class TestMapOrAsync:

        @staticmethod
        async def _helper(v: int) -> str:
            return str(v)

        @pytest.mark.asyncio
        async def test_something(self) -> None:
            s = await Something(1).map_or_async(self._helper, '5')
            assert s == Something('1')

        @pytest.mark.asyncio
        async def test_nothing(self) -> None:
            s = await Nothing().map_or_async(self._helper, '5')
            assert s == Something('5')

    class TestMapOrElse:

        def test_something(self) -> None:
            assert Something(1).map_or_else(str, lambda: '2') == Something('1')

        def test_nothing(self) -> None:
            assert Nothing().map_or_else(str, lambda: '2') == Something('2')

    class TestMapOrElseAsync:

        @staticmethod
        async def _cb(v: int) -> str:
            return str(v)

        @staticmethod
        async def _fb() -> str:
            return '2'

        @pytest.mark.asyncio
        async def test_something(self) -> None:
            assert await Something(1).map_or_else_async(self._cb, self._fb) == Something('1')

        @pytest.mark.asyncio
        async def test_nothing(self) -> None:
            assert await Nothing().map_or_else_async(self._cb, self._fb) == Something('2')

    class TestGet:

        def test_something(self) -> None:
            assert Something(1).get() == 1

        def test_empty(self) -> None:
            assert Nothing().get() is None

        def test_empty_with_fallback(self) -> None:
            assert cast(Nothing[str], Nothing()).get('foo') == 'foo'

    class TestIsSomething:

        def test_something(self) -> None:
            assert Something(1).is_something()

        def test_nothing(self) -> None:
            assert not Nothing().is_something()

    class TestIsNothing:

        def test_something(self) -> None:
            assert not Something(1).is_nothing()

        def test_nothing(self) -> None:
            assert Nothing().is_nothing()

    class TestUnwrap:

        def test_something(self) -> None:
            assert Something('foo').unwrap() == 'foo'

        def test_nothing(self) -> None:
            with pytest.raises(EmptyMaybeError, match='Attempted to unwrap Nothing'):
                assert Nothing().unwrap()

        def test_nothing_msg(self) -> None:
            msg = 'test message'
            with pytest.raises(EmptyMaybeError, match=msg):
                assert Nothing().unwrap(msg)

    class TestUnwrapOr:

        def test_something(self) -> None:
            assert Something('foo').unwrap_or('bar') == 'foo'

        def test_nothing(self) -> None:
            assert cast('Nothing[str]', Nothing()).unwrap_or('bar') == 'bar'

    class TestUnwrapOrElse:

        def test_something(self) -> None:
            assert Something('foo').unwrap_or_else(lambda: 'bar') == 'foo'

        def test_nothing(self) -> None:
            assert cast('Nothing[str]', Nothing()).unwrap_or_else(lambda: 'bar') == 'bar'

    class TestUnwrapOrElseAsync:

        @pytest.mark.asyncio
        async def test_something(self) -> None:
            async def cb() -> str:
                return 'bar'

            v = await Something('foo').unwrap_or_else_async(cb)
            assert v == 'foo'

        @pytest.mark.asyncio
        async def test_nothing(self) -> None:
            async def cb() -> str:
                return 'bar'

            v = await cast('Nothing[str]', Nothing()).unwrap_or_else_async(cb)
            assert v == 'bar'

    class TestBool:

        def test_something(self) -> None:
            assert Something('foo')

        def test_nothing(self) -> None:
            assert not Nothing()

    class TestAndThen:

        @staticmethod
        @wrap_maybe
        def to_int(v: str) -> int | None:
            try:
                return int(v)
            except ValueError:
                return None

        def test_something_invalid(self) -> None:
            assert Something('foo').and_then(self.to_int) == Nothing()

        def test_something_valid(self) -> None:
            assert Something('1').and_then(self.to_int) == Something(1)

        def test_nothing(self) -> None:
            assert Nothing().and_then(self.to_int) == Nothing()

    class TestAndThenAsync:

        @pytest.mark.asyncio
        async def test_something(self) -> None:
            async def f(v: int) -> Maybe[str]:
                return Something(str(v))

            v = await Something(1).and_then_async(f)
            assert v.unwrap() == '1'

        @pytest.mark.asyncio
        async def test_nothing(self) -> None:
            async def f(v: int) -> Maybe[str]:
                pytest.fail()

            v = await Nothing().and_then_async(f)
            assert v == Nothing()

    class TestOrElse:

        def test_something(self) -> None:
            Something('foo').or_else(lambda: maybe('bar')) == Something('foo')

        def test_nothing(self) -> None:
            cast('Nothing[str]', Nothing()).or_else(lambda: maybe('bar')) == Something('bar')

    class TestOrElseAsync:

        @pytest.mark.asyncio
        async def test_something(self) -> None:
            async def cb() -> Maybe[str]:
                return maybe('bar')

            v = await Something('foo').or_else_async(cb)
            assert v == Something('foo')

        @pytest.mark.asyncio
        async def test_nothing(self) -> None:
            async def cb() -> Maybe[str]:
                return maybe('bar')

            v = await cast('Nothing[str]', Nothing()).or_else_async(cb)
            assert v == Something('bar')

    class TestOkOr:

        def test_something(self) -> None:
            Something('foo').ok_or('bar').unwrap() == 'foo'

        def test_nothing(self) -> None:
            cast('Nothing[str]', Nothing()).ok_or('bar').unwrap_err() == 'bar'

    class TestOkOrElse:

        def test_something(self) -> None:
            Something('foo').ok_or_else(lambda: 'bar').unwrap() == 'foo'

        def test_nothing(self) -> None:
            cast('Nothing[str]', Nothing()).ok_or_else(lambda: 'bar').unwrap_err() == 'bar'

    class TestOkOrElseAsync:

        @pytest.mark.asyncio
        async def test_something(self) -> None:
            async def cb() -> str:
                return 'bar'

            v = await Something('foo').ok_or_else_async(cb)
            assert v.unwrap() == 'foo'

        @pytest.mark.asyncio
        async def test_nothing(self) -> None:
            async def cb() -> str:
                return 'bar'

            v = await cast('Nothing[str]', Nothing()).ok_or_else_async(cb)
            assert v.unwrap_err() == 'bar'

    class TestMatch:

        def test_something(self) -> None:
            s = Something('foo')
            match s:
                case Something('bar'):
                    pytest.fail()
                case Something('foo'):
                    assert True
                case Nothing():
                    pytest.fail()
                case _:
                    pytest.fail()

        def test_nothing(self) -> None:
            s: Nothing[str] = Nothing()
            match s:
                case Something('bar'):
                    pytest.fail()
                case Something('foo'):
                    pytest.fail()
                case Nothing():
                    assert True
                case _:
                    pytest.fail()

class TestMaybeFunction:

    def test_something(self) -> None:
        assert maybe('foo') == Something('foo')

    def test_nothing(self) -> None:
        assert maybe(None) == Nothing()


class TestMaybeWrap:

    def test_something(self) -> None:
        @wrap_maybe
        def helper() -> str:
            return 'foo'

        assert helper() == Something('foo')

    def test_nothing(self) -> None:
        @wrap_maybe
        def helper() -> None:
            return None

        assert helper() == Nothing()


class TestMaybeWrapAsync:

    @pytest.mark.asyncio
    async def test_something(self) -> None:
        @wrap_maybe_async
        async def helper() -> str:
            return 'foo'

        assert await helper() == Something('foo')

    @pytest.mark.asyncio
    async def test_nothing(self) -> None:
        @wrap_maybe_async
        async def helper() -> None:
            return None

        assert await helper() == Nothing()


class TestMaybeUnwrap:

    def test_something(self) -> None:
        @unwrap_maybe
        def helper() -> Maybe[str]:
            return Something('foo')

        assert helper() == 'foo'

    def test_nothing(self) -> None:
        @unwrap_maybe
        def helper() -> Maybe[str]:
            return Nothing()

        assert helper() is None


class TestMaybeUnwrapASync:

    @pytest.mark.asyncio
    async def test_something(self) -> None:
        @unwrap_maybe_async
        async def helper() -> Maybe[str]:
            return Something('foo')

        assert await helper() == 'foo'

    @pytest.mark.asyncio
    async def test_nothing(self) -> None:
        @unwrap_maybe_async
        async def helper() -> Maybe[str]:
            return Nothing()

        assert await helper() is None
